"""Main module for executing mc-nag."""

import os
import click
from mc_nag.base_utils.rule_evaluator import RuleEvaluator
from mc_nag.base_utils.click_cli import (
    add_click_options,
    MAIN_OPTIONS
)
from mc_nag.base_utils.models.rule_directory import RuleDirectoryManager
from azure_nag.report_helper import process_templates, report_violations

# pylint: disable=too-many-arguments

STANDARD_RULES_DIR = os.path.realpath(
    f'{os.path.dirname(os.path.abspath(__file__))}/rules'
)


@click.pass_context
def list_rules(ctx, rule_dir_manager):
    """Output information about existing rules."""
    rule_evaluator = RuleEvaluator(None, rule_dir_manager)
    ctx.exit(1 if rule_evaluator.display_rules() else 0)


@click.pass_context
def list_tags_cmd(ctx, rule_dir_manager):
    """Show all the tags for all Rules."""
    rule_evaluator = RuleEvaluator(None, rule_dir_manager)
    rule_evaluator.display_tags()
    ctx.exit(0)


@click.pass_context
def enable_tags_only_cmd(ctx, rule_dir_manager):  # pylint:disable=W0613
    """Show all the tags for all Rules."""
    ctx.exit(0)


@click.command()
@add_click_options(MAIN_OPTIONS)
def main(**kwargs):
    """Perform template parsing and rule evaluation."""
    # Validate either --enable-standard-rules is set or a custom platform
    # rules dir is given
    if not (kwargs['enable_standard_rules'] or kwargs['custom_platform_rules_dir']):
        raise click.BadParameter(
            "Either '--enable-standard-rules' must be True or " +  # noqa: W504
            "'--custom-platform-rules-dir' must have a value."
        )
    rule_dir_manager = RuleDirectoryManager(kwargs['enable_standard_rules'],
                                            STANDARD_RULES_DIR,
                                            kwargs['custom_platform_rules_dir'])
    if kwargs['rules']:
        list_rules(rule_dir_manager)  # pylint: disable=no-value-for-parameter

    if kwargs['list_tags']:
        list_tags_cmd(rule_dir_manager)  # pylint: disable=no-value-for-parameter

    # Read raw template into string
    if not kwargs['filepath']:
        raise click.BadParameter(
            "--filepath must be supplied to evaluate rules."
        )
    azure_template_models = process_templates(kwargs['filepath'],
                                              kwargs['paramfile'])

    report_violations(azure_template_models,
                      kwargs['output'],
                      rule_dir_manager,
                      kwargs['rule_param'],
                      kwargs['enable_tags_only'],
                      kwargs['verbose'])


if __name__ == "__main__":
    # execute only if run as a script
    main()  # pylint: disable=E1120



"""Azure Resource Object Class."""


class AzureResource:
    """Creates Azure Resource Object."""

    def __init__(self, resource_type, resource_name, properties,
                 raw_string, line_number):
        self.resource_type = resource_type
        self.resource_name = resource_name
        self.properties = properties
        self.raw_string = raw_string
        self.line_number = line_number



"""Tests for azure_storageaccount_encrypted_rule.py."""

from mc_nag.base_utils import read_template_file
from azure_nag.models.azure_template import AzureTemplate
from azure_nag.rules.azure_storageaccount_encrypted_rule import (
    AzureStorageAccountEncryptedRule
)

TEMPLATE_PATH = 'tests/templates'


def evaluate_rule(template_path):
    """Evaluate current rule and return output."""
    template_model = AzureTemplate(template_path)
    rule_object = AzureStorageAccountEncryptedRule(template_model)
    return rule_object.evaluate()


def test_encrypted_storageaccount():
    """Happy Path: Ensure rule does not return any resources."""
    test_file = f'{TEMPLATE_PATH}/azure_storageAccounts/' + \
                'valid_storageAccounts.json'
    violating_resources = evaluate_rule(test_file)
    assert not violating_resources


def test_storageaccount_no_encryption():
    """Sad Path: Ensure rule returns violating resources."""
    test_file = f'{TEMPLATE_PATH}/azure_storageAccounts/' + \
                'invalid_storageAccounts_without_encryption.json'
    violating_resources = evaluate_rule(test_file)

    assert len(violating_resources) == 1
    assert violating_resources[0].resource_name == 'insecurestorageaccount'


def test_storageaccount_encryption_service_disabled():
    """Sad Path: Ensure rule returns violating resources."""
    test_file = f'{TEMPLATE_PATH}/azure_storageAccounts/' + \
                'invalid_storageAccounts_encryption_services_disabled.json'
    violating_resources = evaluate_rule(test_file)

    assert len(violating_resources) == 1
    assert violating_resources[0].resource_name == 'insecurestorageaccount'


def test_storageaccount_encryption_missing_keysource():
    """Sad Path: Ensure rule returns violating resources."""
    test_file = f'{TEMPLATE_PATH}/azure_storageAccounts/' + \
                'invalid_storageAccounts_encryption_missing_keysource.json'
    violating_resources = evaluate_rule(test_file)

    assert len(violating_resources) == 1
    assert violating_resources[0].resource_name == 'insecurestorageaccount'



"""Azure Parameter Object Class."""


class AzureParameter:
    """Creates Azure Parameter Object."""

    def __init__(self, parameter_name,
                 parameter_default_value, parameter_type):
        self.parameter_name = parameter_name
        self.parameter_value = parameter_default_value
        self.parameter_type = parameter_type



"""Rule to ensure all StorageAccount resources are encrypted."""

from mc_nag.base_utils.models.rule import (
    BaseRule,
    ERROR as RULE_ERROR
)


class AzureStorageAccountEncryptedRule(BaseRule):
    """Create a rule ensuring the encryption property is created."""

    rule_id = '001'
    description = """Ensure all StorageAccount resources are encrypted by \
validating the 'encryption' property is set and configured."""
    severity = RULE_ERROR
    url = 'https://github.com/stelligent/mc-nag/tests/templates/azure/' + \
          'azure_storageAccounts'
    category_tags = {'storage'}
    source_tags = {'stelligent'}
    resolution = """Add 'encryption' property to the storageAccount resource \
and ensure the service type is set to 'enabled'.

Example:
"encryption": {
  "services": {
    "blob": {
      "enabled": true,
      "keyType": "Account"
    }
  },
  "keySource": "Microsoft.Storage"
},"""

    def evaluate(self):
        """Find storageAccount resources and validate 'encryption' property.

        1. Obtain a list of storageAccount resources.
        2. Search their properties for 'encryption'.
        3. Ensure at least one service is enabled.
        4. Ensure keySource subproperty is set.
        """
        violating_resources = []

        # Obtain list of storageAccounts resources
        storage_accounts = [
            res for res in self.template_model.resources
            if res.resource_type == 'Microsoft.Storage/storageAccounts'
        ]

        # Iterate over storageAccount resources
        for resource in storage_accounts:
            try:
                # Try to access keySource property
                _ = resource.properties['encryption']['keySource']

                # Iterate over services subproperty to see if any are enabled
                services = resource.properties['encryption']['services']
                for _, service_dict in services.items():
                    if 'enabled' not in service_dict or \
                            not service_dict['enabled']:
                        violating_resources.append(resource)
            # Add to violating_resources if keys do not exist
            except KeyError:
                violating_resources.append(resource)

        return violating_resources



"""Report and Output generator helper."""

import os
import sys
from mc_nag.base_utils.rule_evaluator import RuleEvaluator
from mc_nag.base_utils.printers import OutputPrinter, BCOLORS
from azure_nag.models.azure_template import AzureTemplate


def process_templates(filepath, paramfile=None):
    """Processes templates in provided file path."""
    azure_template_models = []
    if os.path.isdir(filepath):
        for root, _, files in os.walk(filepath):
            for file in files:
                azure_template_models.append(AzureTemplate(f'{root}/{file}'))
    else:
        azure_template_models.append(AzureTemplate(filepath, paramfile))
    return azure_template_models


def report_violations(azure_template_models,
                      output,
                      rule_dir_manager,
                      rule_param,
                      enable_tags_only,
                      verbose=0):
    """Helps generate output report from violations."""
    error_violation_counts = 0
    rules_evaluated = []
    for atm in azure_template_models:
        # Instantiate RuleEvaluator and evaluate rules
        rule_evaluator = RuleEvaluator(atm,
                                       rule_dir_manager,
                                       rule_param,
                                       enable_tags_only)
        violations_dict, violations_counts, rules = rule_evaluator.evaluate_rules()
        error_violation_counts += violations_counts.get('ERROR', 0)
        for rule in rules:
            rules_evaluated.append(rule)
        # Print violations in plain text
        print('')
        print(f'-- Template Evaluated: {atm.template_path} --')
        if verbose > 0:
            print(atm)
        getattr(OutputPrinter, output)(violations_dict, violations_counts, rules)
        print()

    if error_violation_counts > 0:
        print_total_report_and_rules(rules_evaluated)
        sys.exit(f'- {BCOLORS["ERROR"]}Total Error Violations{BCOLORS["ENDC"]}: '
                 f'{error_violation_counts}')
    else:
        sys.exit(0)


def print_total_report_and_rules(rules_evaluated):
    """Prints final portion of output with total error violations and rules evaluated."""
    rule_set = [rule.__name__ for rule in sorted(rules_evaluated,
                                                 key=lambda x: x.__name__)]
    print(''.rjust(34, '='))
    print(f'{BCOLORS["UNDERLINE"]}{BCOLORS["HEADER"]}FINAL REPORT{BCOLORS["ENDC"]}')
    print()
    print(f'- {BCOLORS["CYAN"]}Rules Evaluated{BCOLORS["ENDC"]}:')
    for rule in set(rule_set):
        print(f'   {rule}')
    print()



"""AzureParser class."""

from mc_nag.base_utils.parser import BaseParser
from azure_nag.azure_parser_interpolator import AzureParserInterpolator


class AzureParser(BaseParser):
    """Parser class to interpret raw Azure Resource Manager templates."""

    # Dictionary of RegEx patterns that can be picked up in an ARM template
    # resource's property and interpolated
    REGEX_DICT = {
        # Example for below: "[parameters('secret_url')]"
        "PARAMETER_REGEX":
            r"^\[parameters\(\'(.*?)\'\)\]$",

        # Example for below: "[variables('osType')]"
        "VARIABLE_REGEX":
            r"^\[variables\(\'(.*?)\'\)\]$",

        # Example for below: "[variables('environmentSettings')[parameters('environmentName')]]"
        "VARIABLE_PARAMETER_REGEX":
            r"^\[variables\(\'(.*?)\'\)\[parameters\(\'(.*?)\'\)\]\]$",

        # Example for below:
        # "[variables('environmentSettings')[parameters('alternateEnvironmentName')].instanceSize]"
        "VARIABLE_PARAMETER_NESTED_KEY_REGEX":
            r"^\[variables\(\'(.*?)\'\)\[parameters\(\'(.*?)\'\)\]\.(.*?)\]$",

        # Example for below: "[variables('disk_iops_read_write').iops]"
        "VARIABLE_NESTED_KEY_REGEX":
            r"^\[variables\(\'(.*?)\'\)\.(.*?)\]$"
    }

    TEMPLATE_REQUIRED_KEYS = ['$schema', 'contentVersion', 'resources']
    PARAMETERS_FILE_REQUIRED_KEYS = ['$schema', 'contentVersion', 'parameters']

    def validate_file(self,
                      file_type=None,
                      parsed_contents=None,
                      required_keys=None,
                      exception_class=Exception):  # pylint: disable=W0613
        """Validate Azure Resource Manager template or Parameters file has required keys.

        :param file_type: One of 'Template' or 'Parameter'.
        :param parsed_contents: Actual contents of file.
        :param required_keys: List of keys required in the file.
        :param exception_class: Type of exception to raise in case of emergency.
        """
        if file_type == 'Template':
            super().validate_file(file_type='Template',
                                  parsed_contents=self.parsed_template,
                                  required_keys=AzureParser.TEMPLATE_REQUIRED_KEYS)
        elif file_type == 'Parameter':
            if parsed_contents is not None:
                super().validate_file(file_type='Parameter',
                                      parsed_contents=self.parsed_parameter_file,
                                      required_keys=AzureParser.PARAMETERS_FILE_REQUIRED_KEYS)
            else:
                pass

    @staticmethod
    def resolve_value(prop, params_list, variables_list, re_dict):
        """Attempt to resolve a property's value.

        Checks to see if the property is a parameter/variable reference and if so, then attempts
        to resolve that property's value by replacing it with constructed value from the
        reference using RegEx logic. If the property is not a not reference, then just
        returns the raw value of that property.

        :param prop: Property name to resolve.
        :param params_list: List of parameters from which to pull values.
        :param variables_list: List of variables from which to pull values.
        :param re_dict: Dictionary of regex patterns to determine logic to execute.
        """
        for key, _ in re_dict.items():
            if AzureParserInterpolator.get_regex_groups(prop, re_dict, key):
                prop_re_groups = AzureParserInterpolator.get_regex_groups(prop, re_dict, key)
                regex_method = getattr(AzureParserInterpolator, f'resolve_{key.lower()}')
                resolved_property = regex_method(prop_re_groups, variables_list, params_list)
                if resolved_property is not None:
                    return resolved_property
        return prop

    @staticmethod
    def resolve_values_in_properties(res_properties, params_list, variables_list, re_dict):
        """Attempt to resolve a resource's set of property values.

        Iterates through a resource's set of properties and attempts to resolve each
        property's value that is a reference to a parameter/variable.

        :param res_properties: One or more properties from the resource to resolve.
        :param params_list: List of parameters from which to pull values.
        :param variables_list: List of variables from which to pull values.
        :param re_dict: Dictionary of regex patterns to determine logic to execute.
        """
        # Return resolved value for a string
        if isinstance(res_properties, str):
            return AzureParser.resolve_value(res_properties, params_list,
                                             variables_list, re_dict)

        # Iterate over items in list and resolve each value
        if isinstance(res_properties, list):
            return [
                AzureParser.resolve_values_in_properties(
                    item,
                    params_list,
                    variables_list,
                    re_dict)
                for item in res_properties
            ]

        # Iterate over items in dict and resolve each nested value
        if isinstance(res_properties, dict):
            resolved_properties = {}

            for key, value in res_properties.items():
                resolved_properties[key] = AzureParser.resolve_values_in_properties(
                    value, params_list, variables_list, re_dict)

            return resolved_properties

        # Return original value if type not handled
        return res_properties



"""Tests for azure_storageaccount_encrypted_rule.py."""

from mc_nag.base_utils import read_template_file
from azure_nag.models.azure_template import AzureTemplate
from azure_nag.rules.azure_compute_disk_rule import (
    AzureComputeDiskEncryptedRule
)

TEMPLATE_PATH = 'tests/templates/azure_compute_disk'
PARAM_FILES_PATH = f'{TEMPLATE_PATH}/param_files'


def evaluate_rule(template_path, param_file_path=None):
    """Evaluate current rule and return output."""
    if param_file_path is None:
        template_model = AzureTemplate(template_path)
    else:
        template_model = AzureTemplate(template_path, param_file_path)
    rule_object = AzureComputeDiskEncryptedRule(template_model)
    return rule_object.evaluate()
#
#
def test_encrypted_azure_compute_disk_no_param_file():
    """Happy Path: Ensure rule does not return any resources."""
    test_file = f'{TEMPLATE_PATH}/' + \
                'valid_azure_compute_disk.json'
    violating_resources = evaluate_rule(test_file)
    assert not violating_resources

def test_encrypted_azure_compute_disk_with_param_file():
    """Happy Path: Ensure rule does not return any resources."""
    test_file = f'{TEMPLATE_PATH}/' + \
                'valid_azure_compute_disk.json'
    param_file = f'{PARAM_FILES_PATH}/' + \
                'valid_params_file.json'
    violating_resources = evaluate_rule(test_file, param_file)
    assert not violating_resources

def test_not_encrypted_azure_compute_disk_no_param_file():
    """Sad Path: Ensure rule returns violating resources."""
    test_file = f'{TEMPLATE_PATH}/' + \
                'violating_azure_compute_disk_1.json'
    violating_resources = evaluate_rule(test_file)
    assert len(violating_resources) == 1
    assert violating_resources[0].resource_name == 'azure-compute-disk'

def test_not_encrypted_azure_compute_disk_with_param_file():
    """Sad Path: Ensure rule returns violating resources."""
    test_file = f'{TEMPLATE_PATH}/' + \
                'violating_azure_compute_disk_1.json'
    param_file = f'{PARAM_FILES_PATH}/' + \
                 'violating_encryption_params_file.json'
    violating_resources = evaluate_rule(test_file, param_file)
    assert len(violating_resources) == 1
    assert violating_resources[0].resource_name == 'azure-compute-disk-param-test'

def test_azure_compute_disk_no_encryptionsettingscollection_property():
    """Sad Path: Ensure rule returns violating resources."""
    test_file = f'{TEMPLATE_PATH}/' + \
                'violating_azure_compute_disk_2.json'
    violating_resources = evaluate_rule(test_file)
    assert len(violating_resources) == 1
    assert violating_resources[0].resource_name == 'azure-compute-disk'



"""Describe an AzureTemplate in Python memory."""

from mc_nag.base_utils.models.template import BaseTemplate
from mc_nag.base_utils import read_template_file
from ..azure_parser import AzureParser
from .azure_resource import AzureResource
from .azure_parameter import AzureParameter
from .azure_output import AzureOutput
from .azure_function import AzureFunction
from .azure_variable import AzureVariable
from .parameter_helper import check_params_type, \
    check_template_params_default_value, \
    check_file_params_value, \
    check_default_value

# pylint: disable=too-many-instance-attributes

class AzureTemplate(BaseTemplate):
    """Class to model an Azure Resource Manager template."""

    def __init__(self, template_path, parameter_file=None):
        """Initialize AzureObjectModel object."""
        self.template_path = template_path
        self.template_string = read_template_file(template_path)
        # self.parameters_string = None
        # if parameter_file is not None:
        self.parameters_string = read_template_file(parameter_file) \
            if parameter_file is not None else None
        self.parsed_template = AzureParser(self.template_string, self.parameters_string)
        self.parsed_parameter_file = self.parsed_template.parsed_parameter_file
        self._outputs = self._compute_outputs()
        self._functions = self._compute_functions()
        self._variables = self._compute_variables()
        self._parameters = self._compute_parameters()
        self._resources = self._compute_resources()

    @property
    def resources(self):
        """Return computed resources."""
        return self._resources

    @property
    def parameters(self):
        """Return computed parameters."""
        return self._parameters

    @property
    def functions(self):
        """Return computed functions."""
        return self._functions

    @property
    def outputs(self):
        """Return computed outputs."""
        return self._outputs

    @property
    def variables(self):
        """Return computed variables."""
        return self._variables

    def _compute_resources(self):
        """Parse the resources section of Azure Resource Manager templates.

        Create a list of AzureResource to query by other objects during rule
        evaluation.
        Alternate method below for creating dict of resources with key names
        --------------------------------------------------------------------
        def construct_resources_model(self):
            for res in self.parsed_template.resources:
                resource_object = AzureResource(
                    res['type'], res['name'], res['properties'], str(res)
                )
                resource_name = res['name']
                self.resources[resource_name] = {
                    "resource_name": resource_object.resource_name,
                    "resource_type": resource_object.resource_type,
                    "properties": resource_object.properties,
                    "raw_string": resource_object.raw_string
                }
        return: list of AzureResource
        """
        return [
            AzureResource(
                res['type'],
                self.parsed_template.resolve_value(
                    res['name'], self.parameters,
                    self.variables,
                    self.parsed_template.REGEX_DICT),
                self.parsed_template.resolve_values_in_properties(
                    res['properties'],
                    self.parameters,
                    self.variables,
                    self.parsed_template.REGEX_DICT),
                str(res),
                self.parsed_template.find_line_number(
                    f"\"name\": \"{res['name']}\""))
            for res in self.parsed_template.resources]

    def _compute_parameters(self):
        """Create a list of parameters to query by other objects."""
        check_params_type(self.parsed_template.parameters)
        if self.parsed_parameter_file is None:  # pylint: disable=R1705
            check_template_params_default_value(self.parsed_template.parameters)
            return [
                AzureParameter(
                    param_key,
                    param_value['defaultValue'],
                    param_value['type']) for param_key,
                param_value in self.parsed_template.parameters.items()]
        else:
            parameters_list = []
            file_params = self.parsed_parameter_file['parameters']
            check_file_params_value(file_params)
            for param_key, param_value in self.parsed_template.parameters.items():
                if param_key in file_params:
                    resolved_value = file_params[param_key]['value']
                else:
                    check_default_value(param_key, param_value)
                    resolved_value = param_value['defaultValue']

                parameters_list.append(AzureParameter(param_key,
                                                      resolved_value,
                                                      param_value['type']))
            return parameters_list

    def _compute_functions(self):
        """Create a list of functions to query by other objects.

        return: list of AzureFunction
        """
        return [AzureFunction(func['namespace'], func['members'])
                for func in self.parsed_template.functions]

    def _compute_outputs(self):
        """Create a list of outputs to query by other objects.

        return: list of AzureOutput
        """
        return [
            AzureOutput(
                output_key,
                output_value['type'],
                output_value['value']) for output_key,
            output_value in self.parsed_template.outputs.items()]

    def _compute_variables(self):
        """Create a list of variables to query by other objects.

        return: list of AzureVariable
        """
        return [
            AzureVariable(
                variable_key,
                variable_value) for variable_key,
            variable_value in self.parsed_template.variables.items()]



"""Azure Output Object Class."""


class AzureOutput:
    """Creates Azure Output Object."""

    def __init__(self, output_name, output_type, output_value):
        self.output_name = output_name
        self.output_type = output_type
        self.output_value = output_value



"""Rule to ensure all Azure Compute Disk resources are encrypted."""

from mc_nag.base_utils.models.rule import (
    BaseRule,
    ERROR as RULE_ERROR
)


class AzureComputeDiskEncryptedRule(BaseRule):
    """Create a rule ensuring the encryptionSettingsCollection property has 'enabled' as true."""

    rule_id = '002'
    description = """Ensure all AzureComputeDisk resources are encrypted by \
validating the 'encryptionSettingsCollection' property is set and configured \
with the 'enabled' sub-property set to true."""
    severity = RULE_ERROR
    url = 'https://github.com/stelligent/mc-nag/tests/templates/azure/' + \
          'azure_compute_disk'
    category_tags = {'compute-disk'}
    source_tags = {'stelligent'}
    resolution = """Add 'encryptionSettingsCollection' dictionary property to the \
AzureComputeDisk resource and ensure the 'enabled' key is set to true.

Example:
"encryptionSettingsCollection": {
    "enabled": true,
    ....
},"""

    def evaluate(self):
        """Find azureComputeDisk resources and validate 'encryptionSettingsCollection' property.

        1. Obtain a list of azureComputeDisk resources.
        2. Search their properties for 'encryptionSettingsCollection'.
        3. Ensure that the 'encryptionSettingsCollection' property is defined.
        4. Ensure 'enabled' sub-property is set to true.
        """
        violating_resources = []

        # Obtain list of storageAccounts resources
        azure_compute_disks = [
            res for res in self.template_model.resources
            if res.resource_type == 'Microsoft.Compute/disks'
        ]

        # Iterate over azureComputeDisk resources
        for compute_disk in azure_compute_disks:
            try:
                # Try to access encryptionSettingsCollection property
                _ = compute_disk.properties['encryptionSettingsCollection']

                # Check to see if sub-property 'enabled' is set to true
                enabled_key = compute_disk.properties['encryptionSettingsCollection']['enabled']
                if enabled_key is not True:
                    violating_resources.append(compute_disk)

            # Add to violating_resources if keys do not exist
            except KeyError:
                violating_resources.append(compute_disk)

        return violating_resources



"""Azure Variable Object Class."""


class AzureVariable:
    """Creates Azure Variable Object."""

    def __init__(self, variable_name, variable_value):
        self.variable_name = variable_name
        self.variable_value = variable_value



"""Helper module for error handling during AzureParameter object creation."""
from mc_nag.base_utils.exceptions import InvalidParameter


def check_params_type(parameters):
    """Checks that parameters within the template all have the required 'type' keys."""
    missing_param_required_keys = []
    for key, value in parameters.items():
        if 'type' not in value:
            missing_param_required_keys.append(key)
    if len(missing_param_required_keys) > 0:
        raise InvalidParameter("Error - The following template parameters are missing the "
                               f"required 'type' key: \n {missing_param_required_keys}")


def check_template_params_default_value(parameters):
    """Checks that the parameters within a template all have the 'defaultValue' keys.

    This only runs this check when a parameters JSON file is not being passed in.
    """
    missing_param_required_keys = []
    for key, value in parameters.items():
        if 'defaultValue' not in value:
            missing_param_required_keys.append(key)
    if len(missing_param_required_keys) > 0:
        raise InvalidParameter("Error - The following template parameters are missing "
                               f"'defaultValue' keys and are not being passed in as part "
                               f"of a parameters JSON file: \n {missing_param_required_keys}")


def check_file_params_value(file_parameters):
    """Validates the parameters in a parameters JSON file.

    This checks that all of the parameters inside a parameters JSON file contain the required
    'value' keys.
    """
    missing_param_required_keys = []
    for key, value in file_parameters.items():
        if 'value' not in value:
            missing_param_required_keys.append(key)
    if len(missing_param_required_keys) > 0:
        raise InvalidParameter("Error - The following is the list of parameters"
                               " from the passed in parameters JSON file that are missing "
                               f"the required 'value' key: \n {missing_param_required_keys}")


def check_default_value(param_key, param_value):
    """Checks a single template parameter for the 'defaultValue' key.

    This is used to check a single template parameter for the 'defaultValue' key when a parameters
    JSON file is passed in that does include the template parameter.
    """
    if 'defaultValue' not in param_value:
        message = f"Parameter '{param_key}' is missing the 'defaultValue' key inside " \
                  f"the template and was not passed in as part of a " \
                  f"parameters JSON file."
        raise InvalidParameter(message)



"""Azure Function Object Class."""


class AzureFunction:
    """Creates Azure Function Object."""

    def __init__(self, function_namespace, function_members):
        self.function_namespace = function_namespace
        self.function_members = function_members



"""AzureParser RegEx interpolation helper class."""
import re
from mc_nag.base_utils.exceptions import VariableKeyError


class AzureParserInterpolator:
    """AzureParser Interpolator to resolve ARM template dynamic values."""

    @staticmethod
    def check_variables_list(re_var, vars_list):
        """Checks that requested variable exists within the templates Variables."""
        var_names_list = [var.variable_name for var in vars_list]
        if re_var not in var_names_list:
            raise VariableKeyError(f"Variable '{re_var}' does not exist in Variables list!")

    @staticmethod
    def get_regex_groups(prop, re_dict, re_pattern):
        """Evaluates to true/false if the property's value matches the specified RegEx pattern.

        Also retrieves the RegEx groups that were found from the specified pattern.
        """
        return re.search(re_dict[re_pattern], prop)

    @staticmethod
    def resolve_parameter_regex(prop_re_groups, _=None, params_list=None):
        r"""Attempts to resolve the value for a resource property that references a parameter.

        Uses a RegEx pattern to discover the parameter reference and then resolves the value.

        Example resource property value: "[parameters('secret_url')]"
        Example RegEx pattern: r"^\[parameters\(\'(.*?)\'\)\]$"
        """
        for param in params_list:
            if param.parameter_name == prop_re_groups.group(1):
                if param.parameter_value is not None:
                    return param.parameter_value
        return None

    @staticmethod
    def resolve_variable_regex(prop_re_groups, vars_list=None, _=None):
        r"""Attempts to resolve the value for a resource property that references a variable.

        Uses a RegEx pattern to discover the variable reference and then resolves the value.

        Example resource property value: "[variables('osType')]"
        Example RegEx pattern: r"^\[variables\(\'(.*?)\'\)\]$"
        """
        re_var_name = prop_re_groups.group(1)
        AzureParserInterpolator.check_variables_list(re_var_name, vars_list)
        for var in vars_list:
            if var.variable_name == re_var_name:
                if var.variable_value:
                    return var.variable_value
        return None

    @staticmethod
    def resolve_variable_parameter_regex(prop_re_groups, vars_list=None, params_list=None):
        r"""Attempts to resolve the value for a resource property.

        Attempts to resolve the value for a resource property that references a variable which
        accesses a nested key being the value of a parameter.
        Uses a RegEx pattern to discover this specific pattern then resolves the value.

        Example resource property value:
            "[variables('environmentSettings')[parameters('environmentName')]]"
        Example RegEx pattern:
            r"^\[variables\(\'(.*?)\'\)\[parameters\(\'(.*?)\'\)\]\]$"
        """
        re_var_name = prop_re_groups.group(1)
        AzureParserInterpolator.check_variables_list(re_var_name, vars_list)
        param_value = ''
        for param in params_list:
            if param.parameter_name == prop_re_groups.group(2):
                if param.parameter_value:
                    param_value = param.parameter_value
        for var in vars_list:
            if var.variable_name == re_var_name:
                if var.variable_value:
                    var_dict = var.variable_value
                    try:
                        return var_dict[param_value]
                    except KeyError as err:
                        raise VariableKeyError(
                            f"Variable '{var.variable_name}' does not have a key '{err}'.")
        return None

    @staticmethod
    def resolve_variable_parameter_nested_key_regex(
            prop_re_groups, vars_list=None, params_list=None):
        r"""Attempts to resolve the value for a resource property.

        Attempts to resolve the value for a resource property that references a variable which
        accesses a nested key being the value of a parameter and then a deeper nested key.
        Uses a RegEx pattern to discover this specific pattern then resolves the value.

        Example resource property value:
            "[variables('envSettings')[parameters('envName')].instanceSize]"
        Example RegEx pattern:
            r"^\[variables\(\'(.*?)\'\)\[parameters\(\'(.*?)\'\)\]\.(.*?)\]$"
        """
        re_var_name = prop_re_groups.group(1)
        AzureParserInterpolator.check_variables_list(re_var_name, vars_list)
        param_value = ''
        for param in params_list:
            if param.parameter_name == prop_re_groups.group(2):
                if param.parameter_value:
                    param_value = param.parameter_value
        for var in vars_list:
            if var.variable_name == prop_re_groups.group(1):
                if var.variable_value:
                    var_dict = var.variable_value
                    nested_var_key = prop_re_groups.group(3)
                    try:
                        if var_dict[param_value][nested_var_key]:
                            return var_dict[param_value][nested_var_key]
                    except KeyError as err:
                        raise VariableKeyError(
                            f"Variable '{var.variable_name}' "
                            f"does not have a key or nested key '{err}'.")
        return None

    @staticmethod
    def resolve_variable_nested_key_regex(prop_re_groups, vars_list=None, _=None):
        r"""Attempts to resolve the value for a resource property.

        Attempts to resolve the value for a resource property that references a variable which
        accesses a nested key within that variable's dictionary.
        Uses a RegEx pattern to discover this specific pattern then resolves the value.

        Example resource property value:
            "[variables('disk_iops_read_write').iops]"
        Example RegEx pattern:
            r"^\[variables\(\'(.*?)\'\)\.(.*?)\]$"
        """
        re_var_name = prop_re_groups.group(1)
        AzureParserInterpolator.check_variables_list(re_var_name, vars_list)
        for var in vars_list:
            if var.variable_name == re_var_name:
                if var.variable_value:
                    var_dict = var.variable_value
                    nested_var_key = prop_re_groups.group(2)
                    try:
                        if var_dict[nested_var_key]:
                            return var_dict[nested_var_key]
                    except KeyError as err:
                        raise VariableKeyError(
                            f"Variable '{var.variable_name}' does not have a key '{err}'.")
        return None
